/*
This is a part of the mzscript LiteStep module source code.

Copyright (C) 2000-2003 The mzscript Development Team (see mzscript.txt)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include "bangs.h"
#include "mzscript.h"
#include "scripts.h"
#include "variables.h"
#include "utility.h"


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Globals
//
//
extern const TCHAR* g_tzAppName;
extern HWND g_hMain;

BangCommandAssoc g_Bangs[] =
{
    _T("!VarSet"), bangVarSet,
    _T("!VarRemove"), bangVarRemove,

    _T("!VarAdd"), bangVarAdd,
    _T("!VarMul"), bangVarMul,
    _T("!VarMod"), bangVarMod,
    
    _T("!if"), bangIf,

    _T("!exec"), bangExec,
    _T("!msgbox"), bangMsgBox,
    _T("!pause"), bangPause,
    _T("!mzStatus"), bangStatus,
    NULL, NULL
};


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ParseBangParams
//
tstring ParseBangParams(const char* pszArgs)
{
    tstring sReturn;

    if (pszArgs)
    {
        sReturn.assign(pszArgs);
        g_Variables->Expand(sReturn);
    }

    return sReturn;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// bangVarSet
//
// Modifies the value of a variable or adds a new variable
//
void bangVarSet(HWND /* hCaller */, const char* pszArgs)
{
    tstring sVariable;
    tstring sArgs = ParseBangParams(pszArgs);

    if (GetToken(sVariable, sArgs))
    {
        tstring sValue;
        GetToken(sValue, sArgs);

        // TODO: can't set file name
        g_Variables->Set(sVariable, sValue);
    }
    else
    {
        ErrorMessagef(_T("Insufficient parameters for !VarSet.\n"
            "Parameters given: %s"), pszArgs);
    }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// bangVarRemove
//
void bangVarRemove(HWND /* hCaller */, const char* pszArgs)
{
    tstring sVariable;
    tstring sArgs = ParseBangParams(pszArgs);    
    
    if (sArgs.empty())
    {
        ErrorMessagef(_T("Invalid parameters for !VarRemove.\n"
            "Parameters given: %s"), pszArgs);
    }
    else while (GetToken(sVariable, sArgs, true))
    {
        assert(!sVariable.empty());
        g_Variables->Remove(sVariable);
    }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// MathWorker
//
template <class BinaryFunction>
static void MathWorker(LPCSTR pszArgs)
{
    tstring sArgs = ParseBangParams(pszArgs);
    tstring sVariable;
    
    GetToken(sVariable, sArgs);
    if (!sArgs.empty() && GetToken(sArgs, sArgs))
    {
        tstring sValue;
        
        if (!g_Variables->Get(sVariable, &sValue))
        {
            sValue.assign(_T("0"));
        }
        
        sValue = string_cast<int>(BinaryFunction()(atof(sValue.c_str()), atof(sArgs.c_str())));
        g_Variables->Set(sVariable, sValue);
    }
    else
    {
        throw std::invalid_argument("Invalid parameters.\nParameters given: " +
            std::string(pszArgs));
    }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// math !bangs
//
// These try to interpret a variable as an int or double
//
void bangVarAdd(HWND /* hCaller */, const char* pszArgs)
{
    try
    {
        MathWorker<std::plus<int> >(pszArgs);
    }
    catch (std::invalid_argument& ia)
    {
        ErrorMessagef(_T("Error in !VarAdd\n\n%s"), ia.what());
    }
}

void bangVarMul(HWND /* hCaller */, const char* pszArgs)
{
    try
    {
        MathWorker<std::multiplies<int> >(pszArgs);
    }
    catch (std::invalid_argument& ia)
    {
        ErrorMessagef(_T("Error in !VarMul\n\n%s"), ia.what());
    }
}

void bangVarMod(HWND /* hCaller */, const char* pszArgs)
{
    try
    {
        MathWorker<std::modulus<int> >(pszArgs);
    }
    catch (std::invalid_argument& ia)
    {
        ErrorMessagef(_T("Error in !VarMod\n\n%s"), ia.what());
    }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// bangIf
//
void bangIf(HWND /* hCaller */, const char* pszArgs)
{
    tstring sArgs = ParseBangParams(pszArgs);
    
    try
    {
        EvalIf(sArgs);
        ExecCommand(sArgs);
    }
    catch (std::invalid_argument& ia)
    {
        ErrorMessagef(_T("Error during !if execution.\nDescription: %s"),
            ia.what());
    }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// bangExec
//
// Helper !bang to use var expansion outside of scripts
//
void bangExec(HWND /* hCaller */, const char* pszArgs)
{
    tstring sArgs = ParseBangParams(pszArgs);
    
    if (!sArgs.empty())
    {
        ExecCommand(sArgs);
    }
    else
    {
        ErrorMessage(_T("Invalid parameters for !exec.\nNo parameters given."));
    }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// bangMsgBox
//
// Displays a message box
//
void bangMsgBox(HWND /* hCaller */, const char* pszArgs)
{
    tstring sArgs = ParseBangParams(pszArgs);
    
    MessageBox(NULL, sArgs.c_str(), g_tzAppName, MB_TOPMOST | MB_SETFOREGROUND);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// bangPause
//
void bangPause(HWND /* hCaller */, const char* pszArgs)
{
    tstring sArgs = ParseBangParams(pszArgs);

    if (!sArgs.empty())
    {
        unsigned long lInterval = atoi(pszArgs);
        unsigned long lTime = GetTickCount();
        MSG msg;
        
        if( lInterval > 0 )
        {
            while((GetTickCount() - lTime) < lInterval)
            {
                if(PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ))
                {
                    TranslateMessage( &msg );
                    DispatchMessage( &msg );
                }
            }
        }
    }
    else
    {
        ErrorMessage(_T("Invalid parameters for !pause.\nNo parameters given."));
    }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// bangStatus
//
// Temporary about box
//
void bangStatus(HWND /* hCaller */, const char* /* pszArgs */)
{
    TCHAR tzBuffer[MAX_LINE_LENGTH];
    TCHAR tzMessage[MAX_LINE_LENGTH];
    
    SendMessage(g_hMain, LM_GETREVID, 0, (long)&tzBuffer);
    StringCchPrintf(tzMessage, MAX_LINE_LENGTH,
        _T("%s\n\nVariable count:\t%d\nScript count:\t%d"),
        tzBuffer, g_Variables->GetCount(), ScriptCount());
    
    MessageBox(NULL, tzMessage, _T("mzscript Status Report"),
        MB_TOPMOST | MB_SETFOREGROUND | MB_ICONINFORMATION);
}
