/*
This is a part of the mzscript LiteStep module source code.

Copyright (C) 2000-2003 The mzscript Development Team (see mzscript.txt)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include "mzscript.h"
#include "scripts.h"
#include "variables.h"
#include "varfuncs.h"
#include "utility.h"
#include "common.h"


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Globals
//
typedef boost::shared_ptr<AutoWndClass> AutoWndClassPtr;

AutoWndClassPtr g_pAutoWindow;
LineVector g_lvRest;

HWND g_hLitestep = NULL;
HWND g_hMain = NULL;
const UINT g_uMsgs[] = { LM_GETREVID, LM_REFRESH, 0 };

// don't forget to change this in the version resource as well
const TCHAR* g_tzAppName = _T("mzscript");
const TCHAR* g_tzAppVersion = _T("1.0.0");

extern BangCommandAssoc g_Bangs[];


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// initModuleEx
//
int initModuleEx(HWND hLitestep, HINSTANCE hInstance, LPCSTR /* pszPath */)
{
    assert(hLitestep != NULL);
    assert(hInstance != NULL);

    g_hLitestep = hLitestep;

    try
    {
        //
        // Create window, register with LS core
        //
        g_pAutoWindow.reset(new AutoWndClass(g_tzAppName, hInstance,
            MessageHandler));

        g_hMain = g_pAutoWindow->CreateHelperWindow(g_hLitestep);

        SetWindowLong(g_hMain, GWL_USERDATA, magicDWord);
        SendMessage(g_hLitestep, LM_REGISTERMESSAGE,
            (WPARAM)g_hMain, (LPARAM)g_uMsgs);
        
        //
        // Register !bangs, load step.rc settings
        //
        AddBangCommands(g_Bangs);
        
        LoadSetup(false);
    }
    catch (std::runtime_error& re)
    {
        ErrorMessagef(_T("Runtime error during initialization.\n")
                      _T("Description: %s\nPlease contact the module author."),
                      re.what());

        return 1;
    }
    catch (std::bad_alloc&)
    {
        ErrorMessage(_T("Bad allocation during initialization.\n")
                     _T("Out of memory?\n\n")
                     _T("Trying to recover as much as possible..."));
        
        ClearSetup();

        return 2;
    }
    catch (std::exception& e)
    {
        // should never get here...
        ErrorMessagef(_T("Unhandled C++ exception during initialization.\n")
                      _T("Description: %s\nPlease contact the module author."),
                      e.what());

        return 1;
    }
    
    return 0;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// quitModule
//
void quitModule(HINSTANCE /* hInstance */)
{
    RemoveBangCommands(g_Bangs);
    
    SendMessage(g_hLitestep, LM_UNREGISTERMESSAGE,
        (WPARAM)g_hMain, (LPARAM)g_uMsgs);

    ExecCommands(_T("quit"), g_lvRest);
    ClearSetup();

    // normally this isn't necessary since the LS core calls FreeLibrary before
    // calling initModuleEx again, but "better safe than sorry"
    g_pAutoWindow.reset();

    g_hMain = NULL;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// LoadSetup
//
// Things to be loaded both on startup and !refresh
//
void LoadSetup(bool bRefresh)
{
    SetupVarFuncs();
    
    // parse step.rc
    ParseScriptFile(NULL, &g_lvRest);

    ExecCommands(bRefresh ? _T("refresh") : _T("start"), g_lvRest);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ClearSetup
//
// Things to be *un*done both on quit and !refresh
//
void ClearSetup()
{
    g_lvRest.clear();
    ScriptsClear();
    g_Variables->Clear();
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ProcessScript
//
// Process all lines between *Script bang and *Script ~bang
//
static void ProcessScript(FILE* fStart, LPCTSTR pszName)
{
    assert(pszName != NULL); assert(pszName[0]);
    ScriptPtr pScript(new Script(pszName));
    
    bool bCompleted = false;
    TCHAR tzBuffer[MAX_LINE_LENGTH] = { 0 };
    TCHAR tzType[MAX_LINE_LENGTH] = { 0 };
    TCHAR tzRest[MAX_LINE_LENGTH] = { 0 };
    
    // first buffer takes the "*Script" token
    TCHAR* pptzTokens[2] = { NULL, tzType };
    
    while (LCReadNextConfig(fStart, _T("*Script"), tzBuffer, MAX_LINE_LENGTH))
    {
        if (LCTokenize(tzBuffer, pptzTokens, 2, tzRest) >= 2)
        {
            if (lstrcmpi(tzType, _T("~bang")) == 0)
            {
                bCompleted = true;
                break;
            }
            
            pScript->AddLine(tzType, tzRest);
        }
    }

    if (!bCompleted)
    {
        throw ParserError("No corresponding ~bang line");
    }

    ScriptAdd(pScript);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ParseScriptFile
//
bool ParseScriptFile(const TCHAR* ptzFileName, LineVector* plvRest)
{
    bool bReturn = false;
    FILE* f = LCOpen(ptzFileName);
    
    //
    // *mzVarfile
    //
    if (f)
    {
        TCHAR tzBuffer[MAX_LINE_LENGTH] = { 0 };
        TCHAR tzFile[MAX_LINE_LENGTH] = { 0 };
        
        // first buffer takes the "*Script" token
        TCHAR* pptzTokens[2] = { NULL, tzFile };
        
        while (LCReadNextConfig(f, _T("*mzVarfile"), tzBuffer, MAX_LINE_LENGTH))
        {
            if (LCTokenize(tzBuffer, pptzTokens, 2, NULL) >= 2)
            {
                ParseVarFile(tzFile);
            }
            else
            {
                ErrorMessagef(_T("Invalid syntax in *mzVarfile line.\nLine: %s"),
                    tzBuffer);
            }
        }
        
        LCClose(f);
    }
    else
    {
        ErrorMessagef(_T("Error opening \"%s\""),
            ptzFileName ? ptzFileName : _T("step.rc"));
    }

    // due to an lsapi bug/feature the file has to be reopened
    f = LCOpen(ptzFileName);
    
    //
    // *Script
    //
    if (f)
    {
        TCHAR tzBuffer[MAX_LINE_LENGTH] = { 0 };
        TCHAR tzType[MAX_LINE_LENGTH] = { 0 };
        TCHAR tzRest[MAX_LINE_LENGTH] = { 0 };
        
        // first buffer takes the "*Script" token
        TCHAR* pptzTokens[2] = { NULL, tzType };
        
        while (LCReadNextConfig(f, _T("*Script"), tzBuffer, MAX_LINE_LENGTH))
        {
            try
            {
                if (LCTokenize(tzBuffer, pptzTokens, 2, tzRest) >= 2)
                {
                    //
                    // *Script bang
                    //
                    if (lstrcmpi(tzType, _T("bang")) == 0)
                    {
                        // continues reading until ~bang
                        ProcessScript(f, tzRest);
                    }
/*                    //
                    // *Script var (this is more like a *Script const)
                    //
                    else if (lstrcmpi(tzType, _T("var")) == 0)
                    {
                        TCHAR tzName[MAX_LINE_LENGTH] = { 0 };
                        TCHAR tzValue[MAX_LINE_LENGTH] = { 0 };
                        TCHAR* pptzTokens[2] = { tzName, tzValue };
                        
                        if (LCTokenize(tzRest, pptzTokens, 2, NULL) < 2)
                        {
                            throw ParserError("Not enough tokens");
                        }

                        g_Variables->Set(tzName, tzValue);
                    }*/
                    //
                    // further supported *Script lines
                    //
                    else if ((plvRest != NULL) &&
                        ((lstrcmpi(tzType, _T("start")) == 0) ||
                         (lstrcmpi(tzType, _T("refresh")) == 0) ||
                         (lstrcmpi(tzType, _T("quit")) == 0)))
                    {
                        if (!tzRest[0])
                        {
                            throw ParserError("Not enough tokens");
                        }

                        plvRest->push_back(Line(tzType, tzRest));
                    }
                    //
                    // invalid
                    //
                    else
                    {
                        throw ParserError("Invalid line type");
                    }
                }
                else
                {
                    throw ParserError("Not enough tokens");
                }
            }
            catch (ParserError& e)
            {
                if (!e.WasDisplayed() && 
                    (IDNO == AskUser(MB_ICONERROR | MB_YESNO,
                        _T("Error in *Script line.\n\nDescription: %s.\n")
                        _T("Line: %s\n\nContinue parsing?"),
                        e.what(), tzBuffer)))
                {
                    if (plvRest)
                    {
                        plvRest->clear();
                    }
                    
                    break;
                }
            }
        }

        bReturn = true;
        LCClose(f);
    }
    else
    {
        ErrorMessagef(_T("Error opening \"%s\""),
            ptzFileName ? ptzFileName : _T("step.rc"));
    }

    return bReturn;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ParseVarFile
//
void ParseVarFile(const TCHAR* ptzFileName)
{
    FILE* f = LCOpen(ptzFileName);

    if (f)
    {
        TCHAR tzBuffer[MAX_LINE_LENGTH] = { 0 };

        while (LCReadNextLine(f, tzBuffer, MAX_LINE_LENGTH))
        {
            TCHAR tzName[MAX_LINE_LENGTH] = { 0 };
            TCHAR tzValue[MAX_LINE_LENGTH] = { 0 };
            TCHAR* pptzTokens[] = { tzName, tzValue };
            
            // TODO: allow empty variables? ie. check >= 1?
            if (LCTokenize(tzBuffer, pptzTokens, 2, NULL) >= 2 &&
                tzName[0] != _T('*'))
            {
                assert(tzName[0]);
                g_Variables->Set(tzName, tzValue);
            }
        }

        LCClose(f);
    }
    else
    {
        ErrorMessagef(_T("Error opening varfile \"%s\""),
            ptzFileName ? ptzFileName : _T("step.rc"));
    }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ExecCommands
//
void ExecCommands(LPCTSTR pszKey, const LineVector& lvContainer)
{
    assert(pszKey != NULL); assert(pszKey[0]);
    
    for (LineVector::const_iterator iter = lvContainer.begin();
         iter != lvContainer.end(); ++iter)
    {
        if (lstrcmpi(iter->first.c_str(), pszKey) == 0)
        {
            ExecCommand(iter->second);
        }
    }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ExecCommand
//
// Currently just a wrapper around LSExecute, but it could for example look up
// the command in the script !bang table for faster execution.
//
void ExecCommand(const tstring& sCommand)
{
    if(!sCommand.empty())
    {
        LSExecute(NULL, sCommand.c_str(), 0);
    }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// EvalExpression
//
static bool EvalExpression(const tstring& sExpression)
{
    using std::invalid_argument;
    
    //ErrorMessage("Evaluating " + sExpression);
    if (sExpression.empty())
    {
        throw invalid_argument("Empty condition");
    }

    bool bReturn = false;

    // TODO: implement this

/*    // !if [test = 2 && (abc <= 10 || abc >= 20)][!alert yup][!alert nope]
    for (tstring::const_iterator iter = sExpression.begin(); iter != sExpression.end(); ++iter)
    {
        
        if (*iter == _T('\"') || *iter == _T('\''))
        {
            size_t stClose = sExpression.find(*iter, iter - sExpression.begin());
            iter = sExpression.begin() + stClose;
        }
        else if (*iter == _T('('))
        {
            size_t stOpen = iter - sExpression.begin();
            size_t stClose = sExpression.find(_T(')'), stOpen);
            
            if (stClose == tstring::npos)
            {
                throw invalid_argument("No closing parenthesis");
            }

            tstring sSubExpression(sExpression, stOpen+1, stClose-stOpen-1);

            bReturn = EvalExpression(sSubExpression);
            iter = sExpression.begin() + stClose;
            continue;
        }
    }*/

    return bReturn;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// EvalIf
//
bool EvalIf(tstring& sStatement)
{
    using std::invalid_argument;

    bool bReturn = false;
    TCHAR tzCondition[MAX_LINE_LENGTH] = { 0 };
    TCHAR tzTrue[MAX_LINE_LENGTH] = { 0 };
    TCHAR tzFalse[MAX_LINE_LENGTH] = { 0 };

    TCHAR* pptzTokens[] = { tzCondition, tzTrue, tzFalse };

    if (CommandTokenize(sStatement.c_str(), pptzTokens, 3, NULL) >= 2)
    {
        if (!tzCondition[0])
        {
            throw invalid_argument("No condition in if-statement given");
        }

        if (!tzTrue[0] && !tzFalse[0])
        {
            throw invalid_argument("if-statement has neither actions for "
                "\"true\" nor for \"false\"");
        }

        bReturn = EvalExpression(tzCondition);
        sStatement.assign(bReturn ? tzTrue : tzFalse);
    }
    else
    {
        throw invalid_argument("Not enough tokens in if-statement: " +
            sStatement);
    }

    return bReturn;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// MessageHandler
//
LRESULT CALLBACK MessageHandler(HWND hWnd, UINT uMsg, WPARAM wParam,
                                LPARAM lParam)
{
	switch(uMsg)
	{
        case LM_GETREVID:
        {
            if (wParam == 0 || wParam == 1)
            {
                if (SUCCEEDED(StringCchPrintf((TCHAR*)lParam, 64,
                    _T("%s %s"),
                    g_tzAppName, g_tzAppVersion)))
                {
                    return lstrlen((TCHAR*)lParam);
                }
            }
            
            // something above failed
            lParam = NULL;
        }
        break;
        
        case LM_REFRESH:
        {
            ClearSetup();
            LoadSetup(true);
            return 0;
        }
            
    }
    
    return DefWindowProc(hWnd, uMsg, wParam, lParam);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// DllMain
//
// DLL entry point
//
BOOL APIENTRY DllMain(IN HINSTANCE hInstance, IN DWORD dwReason,
                      LPVOID /* pvReserved */)
{
    if (dwReason == DLL_PROCESS_ATTACH)
    {
        DisableThreadLibraryCalls(hInstance);
    }
    
    return TRUE;
}
