/*
This is a part of the mzscript LiteStep module source code.

Copyright (C) 2000-2003 The mzscript Development Team (see mzscript.txt)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include "scripts.h"
#include "mzscript.h"
#include "utility.h"
#include "variables.h"


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// typedefs
//
//
typedef std::vector<ScriptPtr> ScriptVector;


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Local helper functions
//
//
static ScriptVector::iterator ScriptFind(const tstring& sName);
static void ScriptRemove(ScriptVector::iterator iter);


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Globals
//
//
ScriptVector g_Scripts;


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Script::Addline
//
//
void Script::AddLine(LPCTSTR ptzType, const tstring& sRest)
{
    LINETYPE lineType;

    try
    {
        if (lstrcmpi(ptzType, _T("exit")) == 0)
        {
            lineType = EXIT;
        }
        else if (sRest.empty())
        {
            // all linetypes from here on need something in sRest
            throw ParserError("Insufficient tokens");
        }
        else if (lstrcmpi(ptzType, _T("exec")) == 0)
        {
            lineType = EXEC;
        }
        else if (lstrcmpi(ptzType, _T("label")) == 0)
        {
            lineType = LABEL;
        }
        else if (lstrcmpi(ptzType, _T("goto")) == 0)
        {
            lineType = GOTO;
        }
        else if (lstrcmpi(ptzType, _T("gotoif")) == 0)
        {
            lineType = GOTOIF;
        }
        else
        {
            throw ParserError("Invalid line type");
        }
        
        m_Source.push_back(ScriptLine(lineType, sRest));
    }
    catch (ParserError& e)
    {
        if (IDNO == AskUser(MB_ICONERROR | MB_YESNO,
            _T("%s in script \"%s\".\nLine: %s %s\n\n")
            _T("Continue parsing?"),
            e.what(), m_sName.c_str(), ptzType, sRest.c_str()))
        {
            e.SetDisplayed();
            // let the exception travel back to ParseFile()
            throw;
        }
    }
}



//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Script::Execute
//
//
void Script::Execute(Variables::Ptr pVarContext)
{
    assert(pVarContext);
    using std::invalid_argument;
    
    try
    {
        for (ScriptSource::iterator iter = m_Source.begin();
             iter != m_Source.end(); ++iter)
        {
            tstring sLine(iter->second);
            pVarContext->Expand(sLine);

            switch (iter->first)
            {
                case EXEC:
                {
                    ExecCommand(sLine);
                }
                break;
                
                case GOTOIF:
                {
                    EvalIf(sLine);
                    if (sLine.empty())
                    {
                        break;
                    }
                }
                case GOTO:
                {
                    for (iter = m_Source.begin(); iter != m_Source.end(); ++iter)
                    {
                        if (iter->first == LABEL &&
                            lstrcmpi(iter->second.c_str(), sLine.c_str()) == 0)
                        {
                            break;
                        }
                    }
                    
                    if (iter == m_Source.end())
                    {
                        throw invalid_argument("Label \"" + sLine +
                            "\" not found");
                    }
                }
                break;
                
                case EXIT:
                {
                    return;
                }
                break;

                // label lines
                default:
                {
                    assert(iter->first == LABEL);
                    continue;
                }
            }
        }
    }
    catch (invalid_argument& e)
    {
        ErrorMessage(_T("Script execution stopped.\n\nScript: ") + m_sName +
            _T("\nError: ") + e.what());
    }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ScriptFind
//
//
struct ScriptPred
{
public:
    ScriptPred(const tstring& sName):m_sName(sName) {}

    bool operator()(const ScriptPtr pScript) const
    { return (lstrcmpi(m_sName.c_str(), pScript->GetName().c_str()) == 0); }

private:
    const tstring& m_sName;
};

static ScriptVector::iterator ScriptFind(const tstring& sName)
{
    assert(!sName.empty());
    return std::find_if(g_Scripts.begin(), g_Scripts.end(), ScriptPred(sName));
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ScriptAdd
//
//
void ScriptAdd(ScriptPtr pScript)
{
    assert(pScript);
    ScriptVector::iterator iter = ScriptFind(pScript->GetName());
    
    if (iter != g_Scripts.end())
    {
        ScriptRemove(iter);
        ErrorMessage(_T("Script \"") + pScript->GetName() +
            _T("\" already defined. Old version removed."));
    }

    g_Scripts.push_back(pScript);
    AddBangCommandEx(pScript->GetName().c_str(), ScriptBangHandler);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ScriptRemove
//
//
static void ScriptRemove(ScriptVector::iterator iter)
{
    assert(iter != g_Scripts.end());

    RemoveBangCommand((*iter)->GetName().c_str());
    g_Scripts.erase(iter);
}

void ScriptRemove(const tstring& sName)
{
    assert(!sName.empty());
    ScriptVector::iterator iter = ScriptFind(sName);

    if (iter != g_Scripts.end())
    {
        ScriptRemove(iter);
    }
    else
    {
        ErrorMessage(_T("Removal of nonexistent script \"") + sName + _T('\"'));
    }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ScriptsClear
//
//
void ScriptsClear()
{
    g_Scripts.clear();
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ScriptCount
//
//
size_t ScriptCount()
{
    return g_Scripts.size();
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ScriptsBangHandler
//
//
void ScriptBangHandler(HWND /* hCaller */, const char* pszName,
                       const char* pszArgs)
{
    assert(pszName != NULL);
    ScriptVector::iterator iter = ScriptFind(pszName);

    assert(iter != g_Scripts.end());
    assert(*iter);

    Variables::Ptr pLocalVars(new Variables(g_Variables));

    if (pLocalVars)
    {
        Variables::VarPtr pArgsVar = pLocalVars->Set(_T("args"),
            pszArgs ? pszArgs : _T(""));
        
        if (pArgsVar)
        {
            pArgsVar->SetSeparator(_T(' '));
        }
        
        (*iter)->Execute(pLocalVars);
    }
    else
    {
        ErrorMessagef(_T("Couldn't create local variables for Script \"%s\".\n")
                      _T("Script stopped.\n\nOut of memory?"), pszName);
    }
}
