/*
This is a part of the mzscript LiteStep module source code.

Copyright (C) 2000-2003 The mzscript Development Team (see mzscript.txt)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#ifndef _MZSCRIPT_UTILITY_H
#define _MZSCRIPT_UTILITY_H

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Includes
//
#include "common.h"


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// struct BangCommandAssoc
//
// a struct to associate a string with a !bang command function
//
struct BangCommandAssoc
{
    LPCSTR pszName;
    BangCommand* pbCommand;
};


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// case insensitive string comparison (for std::map)
//
struct tstringicmp
{
    bool operator()(IN const tstring& s1, IN const tstring& s2) const
    {
        return (lstrcmpi(s1.c_str(), s2.c_str()) < 0);
    }
};


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Function prototypes
//
// The biggest problem here is the inconsistent mix of LPCTSTR and tstring
//
tstring GetLastErrorStr();
void ReportLastError();
void ReportLastError(LPCTSTR ptzAdditional);
void ErrorMessage(const tstring& sMessage);
void ErrorMessagef(LPCTSTR ptzFormat, ...);

int AskUser(UINT uFlags, LPCTSTR ptzFormat, ...);

void AddBangCommands(const BangCommandAssoc* pBangs);
void RemoveBangCommands(const BangCommandAssoc* pBangs);

bool GetToken(tstring& sToken, tstring& sString, bool bUseBrackets = false);


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// string_cast with specializations
//
/*template <typename Source> inline tstring string_cast(Source source)
{
    return boost::lexical_cast<tstring, Source>(source);
}*/

// the "right" way to do this is to use lexical_cast, but since that uses the
// iostreams we'll stay away from it for now and use an ugly printf hack...
template <typename T> inline tstring string_cast(T source);

template <> inline tstring string_cast(int source)
{
    TCHAR tzBuffer[MAX_PATH];
    StringCchPrintf(tzBuffer, MAX_PATH, _T("%d"), source);
    return tzBuffer;
}

template <> inline tstring string_cast(unsigned short source)
{
    TCHAR tzBuffer[MAX_PATH];
    StringCchPrintf(tzBuffer, MAX_PATH, _T("%u"), source);
    return tzBuffer;
}

template <> inline tstring string_cast(long source)
{
    TCHAR tzBuffer[MAX_PATH];
    StringCchPrintf(tzBuffer, MAX_PATH, _T("%i"), source);
    return tzBuffer;
}

template <> inline tstring string_cast(double source)
{
    TCHAR tzBuffer[MAX_PATH];
    StringCchPrintf(tzBuffer, MAX_PATH, _T("%f"), source);
    return tzBuffer;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// AutoWndClass
//
// Encapsulates window class management, supports a single window
//
class AutoWndClass : boost::noncopyable
{
    // this is meant to be caught as std::runtime_error
    struct winapi_error : public std::runtime_error
    {
        winapi_error() : runtime_error(GetLastErrorStr()) {}
    };
    
public:
    AutoWndClass(const tstring& sClass, HINSTANCE hInst, WNDPROC pfnWndProc) :
      m_hInstance(hInst), m_sClass(sClass), m_hWnd(NULL)
    {
        assert(!IsBadCodePtr((FARPROC)pfnWndProc));
        assert(!sClass.empty());
        assert(hInst != NULL);
        
        WNDCLASS wc;
        ZeroMemory(&wc, sizeof(wc));
        
        wc.lpfnWndProc = pfnWndProc;
        wc.hInstance = hInst;
        wc.lpszClassName = sClass.c_str();
        
        if (!RegisterClass(&wc))
        {
            throw winapi_error();
        }
    }
    
    ~AutoWndClass()
    {
        if (m_hWnd != NULL)
        {
            assert(IsWindow(m_hWnd));
            if (!DestroyWindow(m_hWnd))
            {
                _ReportCleanupError();
            }
        }
        
        if (!UnregisterClass(m_sClass.c_str(), m_hInstance))
        {
            _ReportCleanupError();
        }
    }
    
    HWND CreateHelperWindow(HWND hParent, void* pvParam = NULL)
    {
        assert(m_hWnd == NULL);
        
        m_hWnd = CreateWindowEx(
            WS_EX_TOOLWINDOW | WS_EX_NOPARENTNOTIFY,
            m_sClass.c_str(),
            NULL,
            WS_CHILD | WS_DISABLED,
            0, 0, 0, 0,
            hParent,
            NULL,
            m_hInstance,
            pvParam);
        
        if (!IsWindow(m_hWnd))
        {
            throw winapi_error();
        }
        
        return m_hWnd;
    }
    
private:
    void _ReportCleanupError()
    {
        ReportLastError(_T("Error during cleanup process"));
    }
    
    HINSTANCE m_hInstance;
    tstring m_sClass;
    HWND m_hWnd;
};



#endif // _MZSCRIPT_UTILITY_H