/*
This is a part of the mzscript LiteStep module source code.

Copyright (C) 2000-2003 The mzscript Development Team (see mzscript.txt)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include "variables.h"
#include "utility.h"


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Globals/Static members
//
//    
Variables::Ptr g_Variables(new Variables(Variables::Ptr()));


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Variable::Get
//
tstring Variable::Get(int nIndex) const
{
    if (m_pFunc)
    {
        m_sValue = m_pFunc();
    }
    
    if (nIndex > 0)
    {
        //
        // array handling
        //
        size_t stStart, stEnd;
        if (_GetElementPos(nIndex, stStart, stEnd) == nIndex)
        {
            return m_sValue.substr(stStart, stEnd-stStart);
        }
        else
        {
            return tstring();
        }
    }
    
    return m_sValue;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Variable::Set
//
void Variable::Set(const tstring& sValue, int nIndex)
{
    if (nIndex > 0)
    {
        //
        // array handling
        //
        size_t stStart, stEnd;
        int nElements = _GetElementPos(nIndex, stStart, stEnd);
        
        if (nElements != nIndex)
        {
            m_sValue.append(nIndex-nElements, m_cSeparator);
            stStart = stEnd = m_sValue.length();
        }

        m_sValue.replace(stStart, stEnd-stStart, sValue);
    }
    else
    {
        //
        // change entire value
        //
        m_sValue = sValue;
    }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Variable::Remove
//
// Returns true if the variable can be deleted
//
bool Variable::Remove(int nIndex)
{
    if (nIndex > 0)
    {
        size_t stStart, stEnd;
        
        if (_GetElementPos(nIndex, stStart, stEnd) == nIndex)
        {
            if (nIndex > 1)
            {
                // need to delete the leading separator
                --stStart;
            }
            else if (stEnd != tstring::npos)
            {
                // need to delete the trailing separator
                ++stEnd;
            }

            m_sValue.erase(stStart, stEnd-stStart);
        }

        return false;
    }

    return true;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Variable::_GetElementPos
//
// Get the position of element nIndex in the array
//
int Variable::_GetElementPos(int nIndex, size_t& stStart, size_t& stEnd) const
{
    assert(nIndex > 0);
    /*size_t stLength = m_sValue.length();
    int nCounter = nIndex;
    stEnd = 0;
    
    do
    {
        stStart = stEnd;
        stEnd = m_sValue.find(m_cSeparator, stStart);
    }
    while (--nCounter && stEnd != tstring::npos && ++stEnd < stLength);

    return (nIndex - nCounter);*/

    int nCounter = nIndex;
    bool bInQuote = false;
    tstring::const_iterator iter, itStart, itEnd;

    iter = itEnd = m_sValue.begin();
    itStart = m_sValue.end();
    --itEnd;

    do
    {
        if (iter == m_sValue.end() || (*iter == m_cSeparator && !bInQuote))
        {
            itStart = itEnd;
            itEnd = iter;
            --nCounter;
        }
        else if (*iter == _T('\"'))
        {
            bInQuote = !bInQuote;
        }
    }
    while (iter++ != m_sValue.end() && nCounter);

    stStart = itStart - m_sValue.begin() + 1;
    stEnd = itEnd - m_sValue.begin();

    return (nIndex - nCounter);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Variables::Get
//
bool Variables::Get(tstring sVariable, tstring* psValue) const
{
    assert(!sVariable.empty());
    int nIndex = _InterpretVar(sVariable);

    return _GetValueWorker(sVariable, psValue, nIndex);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Variables::_GetValueWorker
//
bool Variables::_GetValueWorker(const tstring& sVariable, tstring* psValue,
                                int nIndex) const
{
    VarMap::const_iterator iter = m_VarMap.find(sVariable);
    
    if (iter != m_VarMap.end())
    {
        assert(iter->second);
        
        if (psValue)
        {
            *psValue = iter->second->Get(nIndex);
        }
        
        return true;
    }
    
    if (m_pParent)
    {
        return m_pParent->_GetValueWorker(sVariable, psValue, nIndex);
    }
    
    return false;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Variables::Set
//
// Change a variable's value or create a new variable
//
Variables::VarPtr Variables::Set(tstring sVariable, const tstring& sValue,
                                 const tstring& sFile)
{
    assert(!sVariable.empty());
    
    int nIndex = _InterpretVar(sVariable);
    VarMap::iterator iter = m_VarMap.find(sVariable);

    if (iter != m_VarMap.end())
    {
        assert(iter->second);

        // variable already exists, change its value
        iter->second->Set(sValue, nIndex);
        
        return iter->second;
    }
    else
    {
        // create a new variable
        VarPtr pVar = VarPtr(new Variable(sValue, sFile));
        
        if (pVar)
        {
            m_VarMap.insert(VarMap::value_type(sVariable, pVar));
            
            return pVar;;
        }
    }

    ErrorMessage("Couldn't set variable \"" + sVariable + "\" to value \"" +
        sValue + "\".\nOut of memory?");
    
    return VarPtr();
}

bool Variables::Set(const tstring& sVariable, Variable::Func pFunc)
{
    // TODO: implement var saving if var was updated
    assert(!sVariable.empty());
    
    m_VarMap.erase(sVariable);
    
    VarPtr pVar = VarPtr(new Variable(pFunc));
    
    if (pVar)
    {
        m_VarMap.insert(VarMap::value_type(sVariable, pVar));
        return true;
    }
    
    ErrorMessage(_T("Couldn't set function variable \"") + sVariable +
        _T("\".\nOut of memory?"));
    
    return false;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Variables::Remove
//
bool Variables::Remove(tstring sVariable)
{
    assert(!sVariable.empty());
    
    int nIndex = _InterpretVar(sVariable);

    VarMap::iterator iter = m_VarMap.find(sVariable);
    
    if (iter != m_VarMap.end())
    {
        if (iter->second->Remove(nIndex))
        {
            m_VarMap.erase(iter);
        }

        return true;
    }
    
    return false;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// _InterpretVar
//
int Variables::_InterpretVar(tstring& sVariable) const
{
    int nReturn = 0;
    
    size_t stIndex = sVariable.find(_T(':'));
    
    if (stIndex != tstring::npos && sVariable.length() > stIndex+1)
    {
        nReturn = atoi(&sVariable.c_str()[stIndex+1]);
        sVariable.erase(stIndex, sVariable.length()-stIndex);
    }
    
    return nReturn;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Expand
//
bool Variables::Expand(tstring& sString) const
{
    bool bReturn = false;
    
    if (!sString.empty())
    {
        // TODO: check if this is necessary
        // LS $evar$ substitution
        TCHAR tzBuffer[MAX_LINE_LENGTH];
        
        VarExpansion /* Ex */(tzBuffer, sString.c_str()/*, MAX_LINE_LENGTH*/);
        sString.assign(tzBuffer);
        
        // let the worker do the real "work" =)
        bReturn = _ExpandVarsWorker(sString);
    }
    
    return bReturn;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ExpandVarsWorker
//
// Worker function for Expand (see above).
// This is probably rather slow since it uses std::basic_string functions
// instead of c-string manipulation, but it's a lot easier to maintain this way
// and most likely efficient enough for now.
//
bool Variables::_ExpandVarsWorker(tstring& sString) const
{
    const TCHAR cEscape = _T('%');
    
    size_t stPrefix = sString.rfind(cEscape);
    
    // this also catches empty strings
    if (stPrefix == tstring::npos)
    {
        // nothing to expand
        return false;
    }

    if (stPrefix >= sString.length() - 1)
    {
        ErrorMessage(_T("Escape code character found at end of line\n"
            "Expression: ") + sString);

        return false;
    }

    //
    // actual expansion starts here
    //
    switch(sString.at(stPrefix+1))
    {
        //
        // variable/array
        //
        case _T('{'):
        {
            size_t stPostfix = sString.find(_T('}'), stPrefix+2);
            
            if (stPostfix != tstring::npos)
            {
                // Now we know the position of the %{ and the closing }.
                // Construct a new string which holds the variable's name.
                tstring sVariable(sString, stPrefix+2, stPostfix-stPrefix-2);
                tstring sValue;
                
                if (Get(sVariable, &sValue))
                {
                    sString.replace(stPrefix, stPostfix-stPrefix+1, sValue);
                }
                else
                {
                    // replace with empty string if variable doesn't exist
                    sString.erase(stPrefix, stPostfix-stPrefix+1);
                }

            }
            else
            {
                ErrorMessage(_T("Closing } not found.\nExpression: ") +
                    sString);

                return false;
            }
        }
        break;
        
        //
        // escape codes
        //
        case _T('#'):
        {
            sString.replace(stPrefix, 2, _T("$"));
        }
        break;
            
        case _T('='):
        {
            sString.replace(stPrefix, 2, _T("\""));
        }
        break;
            
        default:
        {
            ErrorMessagef(_T("Unknown escape code '%c'.\nExpression: %s"),
                sString.at(stPrefix+1), sString.c_str());
            
            sString.erase(stPrefix, 2);
        }
        break;
    }
    
    // repeat as long as there could be more to expand
    _ExpandVarsWorker(sString);
    
    return true;
}
